#!/usr/bin/perl -w

use strict;
use DBI;

require 'localutils.pl';

my $WAY_BIG = 1000*1000*1000;

###############################
# usage()
###############################

sub usage {
  print "
Usage: $0 [-options <options to pass to createMerge>] beacons_table bumps_table index
";
  exit;
}

###############################
# Okay, real work starts here
###############################

my $options="";

while ( defined($ARGV[0]) and $ARGV[0] =~ /^-/) {
  if ( $ARGV[0] eq '-options' ) {
      shift;
      $options = shift @ARGV;
  }
  else {
      usage();
  }

}
usage() if ( $#ARGV != 2);


my $beaconsTable = shift @ARGV;
my $bumpsTable = shift @ARGV;
my $bIndex = shift @ARGV;

my ($query, $sth);
dbconnect();

# get the number of tables
$query = "SELECT * FROM $beaconsTable"."_TblIndex";
$sth = doQuery( \$query );
my $tblCount = 0;
while ( my ($id, $tblname) = $sth->fetchrow_array ) {
    $tblCount++;
}
my @boundaries;
my @starts = map {0} (1..$tblCount);

$query = "SELECT * from $bumpsTable";
$sth = doQuery(\$query);
while (my $row = $sth->fetchrow_hashref) {

    ##find the immediately previous beacon
    $query = "SELECT * FROM $beaconsTable WHERE id_$bIndex < $$row{packet_id} ORDER BY id DESC limit 1";
    my $presth = doQuery(\$query);
    my $prebeacon = $presth->fetchrow_hashref;

    ##find the immediately after beacon
    $query = "SELECT * FROM $beaconsTable WHERE id_$bIndex >= $$row{packet_id} ORDER BY id limit 1";
    my $poststh = doQuery(\$query);
    my $postbeacon = $poststh->fetchrow_hashref;
    
    my @ends;
    foreach my $index (0..$tblCount-1) {
	if ($index == $bIndex) {
	    push @ends, $$row{packet_id} - 1;
	}
	else {
	    ## this assumes that dividing taking the mid points keeps us in the clear
	    ## i.e., gives us an id that lies in between where the trace is missing
	    push @ends, int($$prebeacon{"id_$index"}/2 + $$postbeacon{"id_$index"}/2); 
	}
    }

    push @boundaries, [join(":", @starts), join(":", @ends)];
    @starts = @ends;  ##createMerge does not do startIds
}
my @ends = map {$WAY_BIG} (1..$tblCount);
push @boundaries, [join(":", @starts), join(":", @ends)];

dbdisconnect();

foreach my $ind (0..$#boundaries) {
    my $cmd = "$options -startids $boundaries[$ind][0] -endids $boundaries[$ind][1] $beaconsTable";
    $cmd = "-boundarycontinue $cmd" if ($ind != 0);
    $cmd = "./createMergedTable.pl $cmd";
    print "$cmd\n";    
}
