#!/usr/bin/perl -w

use strict;

require 'localutils.pl';

my $TETHEREAL = "$ENV{HOME}/usr/local/bin/tethereal";

sub usage {
    print "$0 [-d dir] <machine-name> <interface> <channel>

-d directory; default is '.'
"
}

###parseMDYTime

my %localMonthsHash = (
  Jan => 0,
  Feb => 1,
  Mar => 2,
  Apr => 3,
  May => 4,
  Jun => 5,
  Jul => 6,
  Aug => 7,
  Sep => 8,
  Oct => 9,
  Nov => 10,
  Dec => 11
);

sub parseMDYTime {
  my ($dateStamp) = @_;
  if ( $dateStamp =~ m/([A-Za-z]+) +(\d+), (\d+) (\d+):(\d+):([\d\.]+)/ ) {
    my ($month, $day, $year, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);
    $year = $year - 1900;
    my $m = $localMonthsHash{$month};
    ## timelocal doesn't convert fractional seconds
    my $epochSecs = $sec + timelocal(0, $min, $hour, $day, $m, $year);
    return $epochSecs;
  }
  else {
    die("Couldn't decode date: $dateStamp\n");
  }
}

my $rootDir = ".";
my $verbose = 0;

while (defined $ARGV[0] and $ARGV[0] =~ /^-/) {
    if ( $ARGV[0] eq '-v' ) {
	$verbose = 1;
	shift;
    }
    elsif ( $ARGV[0] eq '-d' ) {
	$rootDir = $ARGV[1];
	shift; shift;
    }
    else {
	usage();
	exit 1;
    }
}

if ( $#ARGV != 2 ) {
    usage();
    exit 1;
}

my $logger = shift @ARGV;
my $interface = shift @ARGV;
my $channel = shift @ARGV;

my @startTimes = ();

dbconnect();

$rootDir =~ s/\/+$//;
my @fileList = split(' ', `ls $rootDir/$logger/small-log*-ath$interface.tcpdump*`);

my $lastresettime = 0;
foreach my $file (sort @fileList) {
    if ($file =~ m/sigcomm-(\d\d\d\d\d\d\d\d-\d\d\d\d)-/) {
	my $time = $1;
	if ($lastresettime ne $time) {
	    push @startTimes, getStartTime($file);
	    $lastresettime = $time;
	}
    }
    else {
	die "unknown file format in $file\n";
    }
}

my $inputtable = "raw_$logger"."_i$interface"."_c$channel";
my $tablename = "bumps_$logger"."_i$interface"."_c$channel";

my $query = "DROP TABLE IF EXISTS $tablename";
doQuery(\$query);

$query = "CREATE TABLE $tablename (id int unsigned auto_increment, packet_id int unsigned, ftime decimal(18,7), primary key(id))";
doQuery(\$query);

$query = "INSERT INTO $tablename (packet_id, ftime) 
             SELECT id, aFrameTime from $inputtable WHERE ".
    join(" OR ", map {"aFrameTime = $_"} @startTimes[1..$#startTimes])." ORDER BY id";

print STDERR "query: $query\n";
my $sth = doQuery(\$query);


dbdisconnect();

sub getStartTime {
    my $file = $_[0];
    
    # Read packets from tethereal verbose output
    die "$TETHEREAL not found\n" if (!-x $TETHEREAL);
    my $command = "$TETHEREAL -V -r $file -o wlan.check_fcs:true";
    open(STREAM, "$command |") or die("Couldn't open packet stream");
    
    my $time;
    while (<STREAM>) {
	if ( s/^\s+Arrival Time:\s+// ) {
	    $time = sprintf("%.7f", &parseMDYTime($_)); 
	    last;
	}
    }
    close(STREAM);
    return $time;
}


