#!/usr/bin/perl -w

# $Id: createMergedTable.pl,v 1.28 2006/08/31 00:57:51 ratul Exp $

use strict;
use DBI;
require 'localutils.pl';

push @INC, "$ENV{HOME}/work/wireless/scripts/DBD";

my $versionFile = '.createMergedTables.version';
my $versionRE = '\..+$';
my $VERSION = q/$Revision: 1.28 $/;  # the string is gen'ed by CVS -- DO NOT EDIT
$VERSION =~ s/^[^\d]+(\d+)\..*$/$1/; # We want only the highest order number of the revision string

my $verbose;
my $LOGFILE;

my $MILLION = 1000*1000;
my $MAX_MTIME = 2**32;
my $ROLLOVER_BOUND_CHECK = 2**31;
my $WAY_BIG = 1000*1000*1000;
my $RingSize = 400; #1000;
my $MERGING_SLACK_USECS = 106.0;
my $ERROR_MERGING_SLACK_USECS = 2.0;
my $BEACON_SKIP_SECS = 1;

## stats on gap briding for merging
my %MergingGap;

my $LIMIT = $WAY_BIG;
my $cumulativeInserts;
my $cumInsertCount = 0;
my $totalInsertCount = 0;

my %addrFields = ("cA1_Dest" => 1, "cA2_Src" => 1, "cA3" => 1, "cA4" => 1);
my @addrFieldsKeys = keys %addrFields;
my %macIndices;

##limiting id range for debugging; make them all zero to switch it off
my %startIds = ( 0 => 0, 1 => 0, 2 => 0, 3 => 0, 4 => 0 );
my %endIds = ( id_0 => $WAY_BIG, id_1 => $WAY_BIG, id_2 => $WAY_BIG );

# this is close to the first common beacon in chihuahuan_sahara_sonoran
#%startIds = ( 0 => 4493*1000, 1 => 4751*1000, 2 => 0 );
#%endIds = ( id_0 => 4503*1000, id_1 => 4761*1000, id_2 => 10*1000 );

# this is where we tripped with closely-spaced packets: 5896581
#%startIds = ( 0 => 5896500, 1 => 1805200);
#%endIds = ( id_0 => $startIds{0}+200, id_1 => $startIds{1}+200);

my @extraFields = ("id", "bRssi"); # "zErr" is added later if we are not ignoring error packets

## Changed mergeVars from array of hash ref to array of array ref,
## for speed.  (jz)
my $PS_INDEX    = 0;          #  per stream variables in mergeNow
my $PS_COUNT    = 1;
my $PS_BUFFER   = 2;

my $TV_STARTID  = 0;          # per table variables in mergeNow
my $TV_ENDID    = 1;
my $TV_LASTTIME = 2;

## Changes details from array of hashes to array of arrays (ratul)
my $DIrollovers = 0;
my $DIstartdelta = 1;
my $DIlastmtime = 2;
my $DIeffectives = 3;

## Changed how beacon table is read, for speed. (jz)
my $BEACONENTRY_id               = 0;
my $BEACONENTRY_dBeaconTimestamp = 1;
my $BEACONENTRY_idX              = 2;
my $BEACONENTRY_ftimeX           = 3;
my $BEACONENTRY_mtimeX           = 4;
my @beaconEntryIndex = (0, 0, [], [], []);  # elements correspond to the indices just defined above

## Also changed how data rows are kept -- arrayref, rather than hashref.
## The below are symbolic names of indices into the data row arrays.
## They are assigned values programmatically.
my $Iid;
my $IaFrameTime;
my $IaTimeDelta;
my $IbChannel;
my $IbHostTime;
my $IbMACTime;
my $IbRate;
my $IbRssi;
my $IbSize;
my $IcA1_Dest;
my $IcA2_Src;
my $IcA3;
my $IcA4;
my $IcDSFrom;
my $IcDSTo;
my $IcDur;
my $IcFragno;
my $IcMoredata;
my $IcMorefrag;
my $IcOrdered;
my $IcProtver;
my $IcPwrmgt;
my $IcRetry;
my $IcSeqno;
my $IcSubtype;
my $IcType;
my $IcWEP;
my $IdBeaconTimestamp;
my $IzErr;
my $IzErrDriver;
my $IzErrPhy;
my $INumStdFields;       # index of the last of these fields

my $IzzFromTableIndex;   # index of the zzFromTable field -- set later

###############################
# usage()
###############################

sub usage {
  print "
Usage: $0 [-v] [-continue] [-boundarycontinue] [-e errThresh] [-s string] beacons_table
        -continue    Continue a previous, aborted execution. (Don't delete existing tables.)
        -e errThresh Set threshold (in usecs.) for combining error packets with any other packet.
                     [Default: 2]
        -dnie        do not ignore error packets (the default is to ignore)
        -l count     Stop after writing (approximately) count rows.  (For testing.)
        -r ringsize  Batch size for select and insert queries.  (Debugging purposes.)
                     [Default: 400]
        -s string    Create table named 'merged_string_...' rather than 'merged_xxx'. (For testing.)
        -v           Verbose (print each SQL query as it's issued)
        -startids    starting ids in the input table; colon separate list
        -endids      starting ids in the input table; colon separate list
        -boundarycontinue  continue on clean boundaries (engineered for kalahari)
        -msu         merging slack in usecs
        -bskips      beacons to skip in seconds 
        -stablemacs  try to keep macindex mapping consistent with the first input table
        -regression  use linear regression to correct timestamps
";

  print "
   This is version $VERSION.  The invocation interface may have changed since
   the last time your ran this program (which was version $_[0]).
   (Your next invocation will succeed, by the way, so be careful!)
" if defined $_[0];
  exit 1;
}


##############################################################
# setupBeaconIndices()
#   Useful only in this merge script, so not put into
#   localutils.pl
##############################################################

sub setupBeaconIndices {
  my ($beaconsTable) = @_;
  my $query = "SHOW COLUMNS FROM $beaconsTable";
  my $sth = doQuery(\$query);
  for (my $index=0;  my ($field) = $sth->fetchrow_array; $index++ ) {
    if    ( $field eq 'id' )               { $beaconEntryIndex[$BEACONENTRY_id] = $index; }
    elsif ( $field eq 'dBeaconTimestamp' ) { $beaconEntryIndex[$BEACONENTRY_dBeaconTimestamp] = $index; }
    elsif ( $field =~ /id_(\d+)/ )         { $beaconEntryIndex[$BEACONENTRY_idX]->[$1] = $index; }
    elsif ( $field =~ /ftime_(\d+)/ )      { $beaconEntryIndex[$BEACONENTRY_ftimeX]->[$1] = $index; }
    elsif ( $field =~ /mtime_(\d+)/ )      { $beaconEntryIndex[$BEACONENTRY_mtimeX]->[$1] = $index; }
  }
}

##############################################################
# checkVersion()
#
#   Because I'm thinking of changing the invocation interface,
#   I'm adding code to alert users that something may have changed.
#   The alert is signaled when the high order field of the cvs 
#   revision is updated.
##############################################################

sub checkVersion {
  system("echo -n 0 >$versionFile") if ! -r $versionFile;
  my $lastVersion = `cat $versionFile | sed 's/$versionRE//'`;
  if ( $lastVersion ne $VERSION ) {
    system( "echo -n '$VERSION' >$versionFile" );
    usage($lastVersion);
  }
}

###############################
# Okay, real work starts here
###############################

#checkVersion();

my $continueOld = 0;
my $boundaryContinue = 0;
my $suffixString;
my $zErrConstraint = " AND zErr IS NULL ";
my $StableMacs = 0;
my $Regression = 0;      ## use linear regression to merge

while ( defined($ARGV[0]) and $ARGV[0] =~ /^-/) {
  if ( $ARGV[0] eq '-v' ) {
    $verbose = 1;
    setVerbose($verbose);
    shift;
  }
  elsif ( $ARGV[0] eq '-continue' ) {
    $continueOld = 1;
    shift;
  }
  elsif ( $ARGV[0] eq '-e' ) {
    shift;
    $ERROR_MERGING_SLACK_USECS = shift;
  }
  elsif ( $ARGV[0] eq '-dnie' ) {
    shift;
    $zErrConstraint = "";
    push @extraFields, "zErr";
  }
  elsif ( $ARGV[0] eq '-l' ) {
    shift;
    $LIMIT = shift;
  }
  elsif ( $ARGV[0] eq '-r' ) {
    shift;
    $RingSize = shift;
  }
  elsif ( $ARGV[0] eq '-s' ) {
    shift;
    $suffixString .= shift;
    $suffixString .= '_' if $suffixString !~ m/_$/;
  }
  elsif ( $ARGV[0] eq '-startids') {
      shift;
      my @ids = split (":", shift);
      map {$startIds{$_} = $ids[$_]} (0..$#ids);
  }
  elsif ( $ARGV[0] eq '-endids') {
      shift;
      my @ids = split (":", shift);
      map {$endIds{"id_$_"} = $ids[$_]} (0..$#ids);
  }
  elsif ($ARGV[0] eq "-boundarycontinue") {
      shift;
      $boundaryContinue = 1;
  }
  elsif ($ARGV[0] eq "-msu") {
      shift;
      $MERGING_SLACK_USECS = shift;
  }
  elsif ($ARGV[0] eq "-bskips") {
      shift;
      $BEACON_SKIP_SECS = shift;
  }
  elsif ($ARGV[0] eq "-stablemacs") {
      shift;
      $StableMacs = 1;
  }
  elsif ($ARGV[0] eq "-regression") {
      shift;
      $Regression = 1;
      $suffixString .= 'reg_';
  }
  else {
      print STDERR "incorrect options\n";
    usage();
  }
}

usage() if ( $#ARGV != 0 );
die "why do you want both continue and boundarycontine?" if ($continueOld and $boundaryContinue);


my $beaconsTable = shift @ARGV;
my $beaconsIndexTable = "$beaconsTable"."_TblIndex";

my $tableName = $beaconsTable;
$tableName =~ s/beacons/merged/;
$tableName =~ s/_/_$suffixString/ if defined $suffixString;

my $tableNameMac = "$tableName"."_MACIndex";

# open the log file
open LOGFILE, ">$tableName.log" || die "ERROR: Couldn't open log file $tableName.log";
$| = 1;        ## do not buffer stdout
my $ofh = select LOGFILE;
$| = 1;        ##do not buffer logfile
select $ofh;
$LOGFILE = *LOGFILE;
setLogFile($LOGFILE);

mirrorPrint("\nStart: ", now(), " tablename = $tableName\n");

dbconnect();

# get the input tables
my $query = "SELECT * FROM $beaconsIndexTable";
my $sth = doQuery( \$query );
my @inputTables;
while ( my ($id, $tblname) = $sth->fetchrow_array ) {
    $tblname = shortenTableName($tblname);
    $inputTables[$id] = $tblname;
}
mirrorPrint("input_tables = ".(join(", ", @inputTables))."\n");

# figure out the name of each of the original, raw data tables used
# to construct the tables being merged (in case they're merge output tables)
my @rawDataTableNames;
my @rawTableOffset;      #starting point
my @rawTableEnd;       #how many raw tables
my @inputTableFromRaw;   #maps raw-table-id to input-table-id 
my @tableZZCount;        # count of the number of per-table zzXXX fields each table requires
my $rawTableCounter = 0;
for ( my $id=0; $id<=$#inputTables; $id++ ) {
  my $mtableIndexName = "$inputTables[$id]_MTblIndex";
  $query = "SHOW TABLE STATUS LIKE '$mtableIndexName'";
  $sth = doQuery(\$query);
  if ( ! $sth->fetchrow_arrayref() ) {
    # table doesnt' exist
    push @rawDataTableNames, ( $inputTables[$id] );
    push @rawTableOffset, ($rawTableCounter);
    push @rawTableEnd, $rawTableCounter;
    push @inputTableFromRaw, $id;
    push @tableZZCount, ($#extraFields+2);
    $rawTableCounter += 1;
  } else {
    $query = "SELECT tablename FROM $mtableIndexName";
    $sth = doQuery(\$query);
    my $count;
    for ($count=0; my ($name) = $sth->fetchrow_array; $count++ ) {
      push @rawDataTableNames, ( $name );
      push @inputTableFromRaw, $id;
    }
    push @rawTableOffset, ($rawTableCounter);
    push @rawTableEnd, $rawTableCounter + $count - 1;
    push @tableZZCount, ( ($#extraFields+2)*$count );
    $rawTableCounter += $count;
  }
}

## read the MACIndexTables
my @inputMACIndices;
readInputMACIndices();

# extract the column names; don't want to hardcode the current table format for raw packets
my @inColumns;
my %outColumnType;
foreach my $id (0..$#inputTables) {
  $query = "SHOW COLUMNS FROM $inputTables[$id]";
  $sth = doQuery(\$query);
  while (my ($field, $type, $null, $key, $default, $extra) = $sth->fetchrow_array) {
    last if $field =~ /^zz/;
    push @{$inColumns[$id]}, $field;
    #	$field .= "_$id" if ($field =~ m/^zz/); ## if its a merged field
    #	$outColumnType{$field} = $type;
    $outColumnType{$field} = $type if ($field !~ m/^zz/); ## if its not a merged field
  }
}

#my @outfieldName = grep {$_ ne "id"} sort keys %outColumnType;  ##comes in handy later
my @outfieldName = @{$inColumns[0]}[1 .. $#{$inColumns[0]}];

## extra columns generated by the merge are prefixed by zz
## Some come from the original table data, but two are "special" -- 
##   synthesized by this script -- zzFromTable and zzTDiff_X

$outColumnType{'zzFromTable'} = 'tinyint UNSIGNED';
push @outfieldName,('zzFromTable');
for (my $id=0; $id<$rawTableCounter; $id++) {
  foreach my $field (@extraFields) {
    my $newField = "zz$field"."_$id";
    $outColumnType{$newField} = $outColumnType{$field};
    push @outfieldName,($newField);
  }
  # zztDiff is handled specially
  $outColumnType{"zzTDiff_$id"} = 'tinyint UNSIGNED';
  push @outfieldName, ("zzTDiff_$id");
}

#print "extraFields: ", join(', ',@extraFields), "\n";
#print "outfieldName: ", join(', ', @outfieldName ), "\n";
#print "outColumnType:\n\t", join("\n\t", (map{ $_ . "\t" . $outColumnType{$_} } sort keys %outColumnType) ), "\n";
#print "tableZZCount: ", join(', ', @tableZZCount), "\n";

# figure out index values for the fields of the beacon table
setupBeaconIndices( $beaconsTable );

# set values for the $Ifieldname index variables
my $index;
for ( $index=0; $index<=$#{$inColumns[0]}; $index++ ) {
  last if $inColumns[0]->[$index] =~ /^zz/;
  my $field = $inColumns[0]->[$index];
  my $cmd = "\$I$field = $index";
  eval( $cmd );
  $INumStdFields = $index;
}
$IzzFromTableIndex = $index;   # zzFromTable MUST BE THE FIRST zzXXX field for this to work

# sanity check
die "ERROR: Index mismatch on id: $Iid vs. 0" if $Iid != 0;
for ( my $index=0; $index<=$INumStdFields-1; $index++ ) {
  my $cmd = "die \"ERROR: Index mismatch on $outfieldName[$index]: \$I$outfieldName[$index] vs. " . ($index+1) . "\" if \$I$outfieldName[$index] != " . ($index+1) . "; 1;";
  eval( $cmd ) ||  die $@;
}

# save indices for addrFields in a more convenient format
my @addrFieldsIndices;
foreach (@addrFieldsKeys) {
  my $cmd = "push \@addrFieldsIndices, (\$I$_)";
  eval( $cmd );
}

# save indices for the extraFields in a more convenient format
my @extraFieldsIndices;
foreach (@extraFields) {
  my $cmd = "push \@extraFieldsIndices,(\$I$_)";
  eval( $cmd );
}
# tack on the special extra field name. @extraFields now has the names of all per-table additional fields.
# (There is also the zzFromTable field, which sits after the raw data fields and the other zzXXX fields.)
push @extraFields, ('zzTDiff');

##time adjustments are going to be in reference to the first input table
my @linearFactors;
$linearFactors[0]->{slope} = 1;
$linearFactors[0]->{intercept} = 0;
$linearFactors[0]->{base} = 0;

#### everything above is independent of any input data
#### the real input-data dependent stuff happens below

## a helper structure
my @details;
initializeDetails();

# a commonly used constraint
my $idConstraint = join (" AND\n", map {"id_$_ IS NOT NULL AND id_$_ <= ".$endIds{"id_$_"}} (0..$#inputTables));

# another helper structure
my $lastBeaconRow;
my $finalBeaconRow;
map {$lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[$_]] = $startIds{$_}} (0..$#inputTables);
map {$finalBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[$_]] = $endIds{"id_$_"}} (0..$#inputTables);
print "finalBeaconRow: ", join(', ', map{defined($_)?$_:"NULL"} @{$finalBeaconRow}), "\n";
print "beaconEntryIndex: ", join (', ', map{ defined($_)?$_:"NULL"} @{$beaconEntryIndex[$BEACONENTRY_idX]}), "\n";
print "merging slack = $MERGING_SLACK_USECS beacon skip = $BEACON_SKIP_SECS\n";

# deal with restarts
my @continueStarts;

if (!$continueOld and !$boundaryContinue) {   #create the new tables
    $query = "DROP TABLE IF EXISTS $tableName";
    doQuery(\$query);
    $query = "CREATE TABLE $tableName ( id int NOT NULL AUTO_INCREMENT,\n";
    map {$query .= "$_ $outColumnType{$_},\n"} (@outfieldName);
    $query .= " PRIMARY KEY( id ) )";
    doQuery(\$query);

    $query = "DROP TABLE IF EXISTS $tableNameMac";
    doQuery(\$query);
    $query = "CREATE TABLE $tableNameMac ( id INT UNSIGNED NOT NULL AUTO_INCREMENT,
                                    value char(18), 
                                    PRIMARY KEY (id))";
    doQuery(\$query);

    my $mtableIndexName = $tableName . '_MTblIndex';
    $query = "DROP TABLE IF EXISTS $mtableIndexName";
    doQuery(\$query);
    $query = "CREATE TABLE $mtableIndexName ( id INT UNSIGNED NOT NULL AUTO_INCREMENT,
                                    tablename char(48),
                                    PRIMARY KEY (id))";
    doQuery(\$query);
    $query = "INSERT INTO $mtableIndexName (tablename) VALUES ('" . join("'),('", @rawDataTableNames) . "')";
    doQuery(\$query);
}
elsif ($continueOld) {
    @continueStarts = initializeContinueState();
}
else  {
    initializeBoundaryContinueState();
}


# start reading rows of beacons table that have all the ids
my $conn = getConnection();
$query = "SELECT * from $beaconsTable 
          WHERE id_0 > $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[0]] AND $idConstraint";
my $sthconn = doQueryConnUse($conn, \$query);

if ($Regression) {
    computeLinearRegression($sthconn, \@linearFactors, $query);
    print STDERR "lfacts for 1: $linearFactors[1]->{slope} $linearFactors[1]->{intercept}\n";
}

my $row;
while (!$Regression && 
       ($row = $sthconn->fetchrow_arrayref) && ($totalInsertCount < $LIMIT)) {

    ##lets not do close beacons
    next if ($lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[0]] != $startIds{0} &&  ## always do the first row
	     $row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[0]] - $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[0]] < $BEACON_SKIP_SECS );

    for my $id (0..$#inputTables) {
	my $originalMtime = 	$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $details[$id]->[$DIrollovers];
	
	
	die "ERROR: couldn't account for all the effectives for $id\n" if ($details[$id]->[$DIeffectives] != 0);
	
	##we might have multiply rolled over since the last beacon; compute the effective mtime
	## based on that the starting delta and the current delta should be similar
	while (abs($row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION -
		   $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] - $details[$id]->[$DIstartdelta])
	       > $ROLLOVER_BOUND_CHECK) {
	    my $derails = "id=$id".
		" beid=".$row->[$beaconEntryIndex[$BEACONENTRY_idX]->[$id]].
		" omtime=".$originalMtime.
		" mtime=".$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]].
		" ro=".$details[$id]->[$DIrollovers]/$MAX_MTIME.
		" ftime=".sprintf("%.7f",$row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION).
		" delta=".$details[$id]->[$DIstartdelta];
	    mirrorPrint("effective $derails\n");
	    
	    $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $MAX_MTIME;
	    die sprintf("ERROR: cannot get effective mtime. ftime = %.7f; mtime = %Ld\n", 
			$row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]],
			$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]])
		if ( $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] > 1 + $row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION);
 
	    $details[$id]->[$DIeffectives]++;
	}
    }
    
    foreach my $id (1..$#inputTables) {
	if ($lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[0]] == $startIds{0}) {  ## this is the first row
	    $linearFactors[$id]->{slope} = 1;
	    $linearFactors[$id]->{base} = 0;
	    $linearFactors[$id]->{intercept} = $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]] -
		$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	}
	else {
	    $linearFactors[$id]->{slope} =  ($row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]] -
					     $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]]) /
					     ($row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] -
					      $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]]);
	    $linearFactors[$id]->{base} = $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	    $linearFactors[$id]->{intercept} = $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]] -
		$lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	}
#	    mirrorPrint("lfacs for ".($row->[$BEACONENTRY_dBeaconTimestamp]).": id=$id  int=".
#		($linearFactors[$id]->{intercept})." slope=".($linearFactors[$id]->{slope})." or ($$row{mtime_0} - $lastBeaconRow->{mtime_0})/(".$$row{"mtime_$id"}." - ".$lastBeaconRow->{"mtime_$id"}.")\n");
    }
    
    &mergeNow($lastBeaconRow, $row);
    $lastBeaconRow = [@{$row}];   # some limited testing indicated that this is the fastest approach to reading
    # rows (fetchrow_arrayref) and then creating copies of them
    last if $totalInsertCount >= $LIMIT;
}

if ( $totalInsertCount < $LIMIT ) {
  # deal with records occuring after final beacon in beacon table
    foreach my $id (0..$#inputTables) {
	if ($Regression) {
	    # do nothing
	}
	elsif (defined($lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]])) {
	    $linearFactors[$id]->{intercept} = $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]] -
		$lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	    $linearFactors[$id]->{base} = $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	    ##we stick with slope from the last time
	}
	else {
	    print STDERR "warning: no beacons found!\n";
	    $linearFactors[$id]->{intercept} = 0;
	    $linearFactors[$id]->{base} = 0;
	    $linearFactors[$id]->{slope} = 1;
	}
    }
    &mergeNow($lastBeaconRow, $finalBeaconRow);
}

# flush any remaining rows from output buffer
insertIntoTable() if ($cumInsertCount > 0); ##left over inserts

insertStableMacs() if ($StableMacs);

printHistogram();

dbdisconnect();
mirrorPrint("\nTerminated normally(?)\n");
mirrorPrint(now(), "\n");

################################

##uses: lastrow and row id's to compute the range to packets to process
sub mergeNow() {
  my ($lastBeaconRow, $row) = @_;

  my $id;
  my $minpktRef;
  my $minid;
  my $minTime;
  my $otherPktRef;
  my $diff;
  my @needZZ;
  my $pktToInsert;
  my @zzFieldArr;
  my $pkt;
  my $index;
  my $bestRSSI;
  my $fromTable;
  my @diffVal;
  my @nextPacket;   ## one from each input table, if there are any left in that table
  my $id1;

  # initialize stream status structure
  my @tableVars;
  my @packetStreams;
  
  foreach $id (0 .. $#inputTables) {
      $tableVars[$id]->[$TV_STARTID] = $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[$id]];
      $tableVars[$id]->[$TV_ENDID] = $row->[$beaconEntryIndex[$BEACONENTRY_idX]->[$id]];
 
      map {$packetStreams[$_]->[$PS_COUNT] = $packetStreams[$_]->[$PS_INDEX] = 0;} ($rawTableOffset[$id] .. $rawTableEnd[$id]);
  }

  # initialize row streams
  foreach $id (0..$#inputTables) {
      readSomeMore($id, $tableVars[$id], \@packetStreams, $linearFactors[$id], $details[$id]);

      foreach ($rawTableOffset[$id] .. $rawTableEnd[$id]) {
	  if ( $packetStreams[$_]->[$PS_COUNT] > 0 ) {
	      $nextPacket[$_] = $packetStreams[$_]->[$PS_BUFFER]->[0];
	      $packetStreams[$_]->[$PS_INDEX] = 1;
	      $packetStreams[$_]->[$PS_COUNT]--;
	  }
	  else {
	      $nextPacket[$_] = undef;
	  }
      }
  }

  while (1) {

    ## 1. find the minimum
    $minpktRef = undef;
    $minTime = $WAY_BIG * 1000 * 1000;
    foreach $id (0..$#packetStreams) {
	if (!defined $nextPacket[$id] and $packetStreams[$id]->[$PS_COUNT]>0 ) {
	    $nextPacket[$id] = $packetStreams[$id]->[$PS_BUFFER]->[$packetStreams[$id]->[$PS_INDEX]];
	    $packetStreams[$id]->[$PS_INDEX]++;
	    $packetStreams[$id]->[$PS_INDEX] = 0 if ($packetStreams[$id]->[$PS_INDEX] == $RingSize);
	    $packetStreams[$id]->[$PS_COUNT]--;
	}
	if ( defined($nextPacket[$id]) && ($nextPacket[$id]->[$IbMACTime] < $minTime) ) {
	    $minpktRef = $nextPacket[$id];
	    $minTime = $minpktRef->[$IbMACTime];
	    $minid = $id;
	}
    }
    last if (!defined $minpktRef);  ##undefined when we run out of all packets

    ## 2. see if the other packets on the top of the list are duplicates and can be merged
    @needZZ = ();
    @diffVal = ();  # store only diffs for streams included in this output packet

    foreach $id (0..$#packetStreams) {
      if ($id == $minid ) {
	push @needZZ, ( $minpktRef );
	push @diffVal,(0);
      }
      elsif ( !defined($nextPacket[$id]) ) {
	push @needZZ, ( undef );
      }
      elsif ( $id >= $rawTableOffset[$inputTableFromRaw[$minid]] and $id <= $rawTableEnd[$inputTableFromRaw[$minid]]) {
	  # no point trying to merge with a packet that originated from the same merged table
	  push @needZZ, ( undef );
      }  
      else {
	$otherPktRef = $nextPacket[$id];
	$diff = $otherPktRef->[$IbMACTime] - $minTime;

	if ($diff >=  $MERGING_SLACK_USECS) {  ##consider only close packets
	  push @needZZ, ( undef );
	}
	else {
#	    print STDERR "$minpktRef->[$Iid] $otherPktRef->[$Iid]\n";
#	    print STDERR "$minpktRef->[$Iid] -$minpktRef->[$IzErr]-    $otherPktRef->[$Iid] -$otherPktRef->[$IzErr]-\n" if ($minpktRef->[$Iid] == 12815);

	  ## now determine duplicacy
	  if (defined($minpktRef->[$IzErr]) and defined($otherPktRef->[$IzErr])) {
	    $MergingGap{"error2"}->{int $diff}++;
	    if ( $diff <= $ERROR_MERGING_SLACK_USECS ) {
	      push @needZZ, ( $otherPktRef );
	      push @diffVal, ( $diff );
	    } else {
	      push @needZZ, ( undef );
	    }
	  }
	  elsif (defined($minpktRef->[$IzErr]) or defined($otherPktRef->[$IzErr])) {
	    $MergingGap{"error1"}->{int $diff}++;
	    if ( $diff <= $ERROR_MERGING_SLACK_USECS ) {
	      push @needZZ, ( $otherPktRef );
	      push @diffVal, ( $diff );
	    } else {
	      push @needZZ, ( undef );
	    }
	  }
	  elsif ( $minpktRef->[$IcA1_Dest] == $otherPktRef->[$IcA1_Dest] &&
		  ( ## either both should be undefined or they should be defined and equal
		   (!defined($minpktRef->[$IcA2_Src]) && !defined($otherPktRef->[$IcA2_Src])) ||
		   (defined($minpktRef->[$IcA2_Src]) && defined($otherPktRef->[$IcA2_Src]) && $minpktRef->[$IcA2_Src] == $otherPktRef->[$IcA2_Src])
		  ) &&
		  ( ## seqno is defined => fragno is defined
		   (!defined($minpktRef->[$IcSeqno]) && !defined($otherPktRef->[$IcSeqno])) ||
		   (defined($minpktRef->[$IcSeqno]) && defined($otherPktRef->[$IcSeqno]) && ($minpktRef->[$IcSeqno] == $otherPktRef->[$IcSeqno]) &&
		    ($minpktRef->[$IcFragno] == $otherPktRef->[$IcFragno]))
		  ) &&
		  $minpktRef->[$IbRate]    == $otherPktRef->[$IbRate] &&
		  $minpktRef->[$IbSize]    == $otherPktRef->[$IbSize] &&
		  $minpktRef->[$IcSubtype] == $otherPktRef->[$IcSubtype] &&
		  $minpktRef->[$IcRetry]   == $otherPktRef->[$IcRetry]
		) {
	    $MergingGap{"goodmatch"}->{int $diff}++;
	    push @needZZ, ( $otherPktRef );
	    push @diffVal, ( $diff );
	  }
	  else {
	    $MergingGap{"almostmatch"}->{int $diff}++;
	    push @needZZ, ( undef );
	  }
	}
      }
    }

    ## 3. get ready to insert packets for this row, and consume them in any case
    $bestRSSI = undef;
    @zzFieldArr = ();

    for ($id=0; $id<=$#needZZ; $id++) {
      $pkt = $needZZ[$id];

      if ( ! defined($pkt) ) {
	  if (!defined $zzFieldArr[$id]) {  ##a merged packet has not already filled this
	      $zzFieldArr[$id] = "";
	      for ($index=0; $index <= $#extraFields; $index++) {
		  $zzFieldArr[$id] .= ",NULL";
	      }
	  }
	next;
      }

      # if the table being merged is raw data, create the zzXXX fields for it.
      # Otherwise, they already exist -- copy them.
      if ( !defined($pkt->[$IzzFromTableIndex]) ) {  # if that field exists, the pkt comes from a merged table
	  $zzFieldArr[$id] = "";
	foreach $index (@extraFieldsIndices) {
	    $zzFieldArr[$id] .= (defined($pkt->[$index])) ? ",$pkt->[$index]" : ",NULL";
	}
	  # add in the special field (zzTDiff)
	  $zzFieldArr[$id] .= ',' . shift @diffVal;
      }
      else {
	# Copy all the per-table zzXXX fields from this packet into the merged table
	# Have to translate the zzTDiff fields first.
	$diff = shift @diffVal;
	for ($index=$IzzFromTableIndex+$#extraFields+1; $index<=$#{$pkt}; $index+=$#extraFields+1) {
	  $pkt->[$index] += $diff if defined($pkt->[$index]);
	}
	$index=$IzzFromTableIndex+1;  # a merged packet speaks for all its constituent raw tables
	for ($id1=$rawTableOffset[$inputTableFromRaw[$id]]; $id1 <= $rawTableEnd[$inputTableFromRaw[$id]]; $id1++) {
	    $zzFieldArr[$id1] = ""; 
	    for (0..$#extraFields) {
		$zzFieldArr[$id1] .= defined($pkt->[$index]) ? ",$pkt->[$index]" : ",NULL";
		$index++;
	    }
	}
      }

      $bestRSSI = $pkt->[$IbRssi] if !defined($bestRSSI) or (defined($pkt->[$IbRssi]) && $pkt->[$IbRssi] > $bestRSSI);

      ##take the other packet's fields if minpkt is error'd
      if ( defined($minpktRef->[$IzErr]) ) {
	$minpktRef = $pkt;
	$minid = $id;
      }

      $nextPacket[$id] = undef;

      if ( $packetStreams[$id]->[$PS_COUNT] == 0 and 
	   $pkt->[$IaFrameTime] < $tableVars[$inputTableFromRaw[$id]]->[$TV_LASTTIME] +  2 * $MERGING_SLACK_USECS) {
	  readSomeMore($inputTableFromRaw[$id], $tableVars[$inputTableFromRaw[$id]], \@packetStreams, 
		       $linearFactors[$inputTableFromRaw[$id]], $details[$inputTableFromRaw[$id]]);
      }

#       # update the head-of-line packet for this stream
#       if ( $packetStreams[$id]->[$PS_COUNT]>0 ) {
# 	$nextPacket[$id] = $packetStreams[$id]->[$PS_BUFFER]->[$packetStreams[$id]->[$PS_INDEX]];
# 	$packetStreams[$id]->[$PS_INDEX]++;
# 	$packetStreams[$id]->[$PS_INDEX] = 0 if ($packetStreams[$id]->[$PS_INDEX] == $RingSize);
# 	$packetStreams[$id]->[$PS_COUNT]--;
#       }
#       ##read more only if we do not have sufficient look ahead in buffers from that table
#       elsif ($pkt->[$IaFrameTime] < $tableVars[$inputTableFromRaw[$id]]->[$TV_LASTTIME] +  2 * $MERGING_SLACK_USECS) {
# 	readSomeMore($inputTableFromRaw[$id], $tableVars[$inputTableFromRaw[$id]], \@packetStreams, 
# 		     $linearFactors[$inputTableFromRaw[$id]], $details[$inputTableFromRaw[$id]]);
# 	if ( $packetStreams[$id]->[$PS_COUNT] > 0 ) {
# 	    $nextPacket[$id] = $packetStreams[$id]->[$PS_BUFFER]->[$packetStreams[$id]->[$PS_INDEX]];
# 	    $packetStreams[$id]->[$PS_INDEX]++;
# 	    $packetStreams[$id]->[$PS_INDEX] = 0 if ($packetStreams[$id]->[$PS_INDEX] == $RingSize);
# 	    $packetStreams[$id]->[$PS_COUNT]--;
# 	} else {
# 	    $nextPacket[$id] = undef;
# 	}
#       }
#       else {
# 	  $nextPacket[$id] = undef;
#       }
  
  }

    # 4. write rows, unless still restarting
    #if (!$continueOld or $minpktRef->[$Izzid_X[$minid]] > $continueStarts[$minid]) {
    if (!$continueOld or $minpktRef->[0] > $continueStarts[$inputTableFromRaw[$minid]]) {
      print "restarting: $minid $inputTableFromRaw[$minid]  $minpktRef->[0]  $continueStarts[$inputTableFromRaw[$minid]]\n" if $continueOld;
      $continueOld = 0;

      ## keep the best RSSI
      $minpktRef->[$IbRssi] = $bestRSSI;

      #fix the time fields
      $minpktRef->[$IaFrameTime] = sprintf("%.7f", ($details[0]->[$DIstartdelta] + $minpktRef->[$IbMACTime])/$MILLION);
      $minpktRef->[$IbMACTime] = ($details[0]->[$DIstartdelta] + $minpktRef->[$IbMACTime]) % $MAX_MTIME;

      # skip the packet's id field
      $pktToInsert = (defined $minpktRef->[1]) ? $minpktRef->[1] : "NULL";
      map {$pktToInsert .= (defined $minpktRef->[$_] )? ",$minpktRef->[$_]" : ",NULL" } (2..$INumStdFields);


      $fromTable = 1 + $minid;

      # append other zzFields

      $cumulativeInserts .= ",($pktToInsert,$fromTable@zzFieldArr)";
      $cumInsertCount++;
      $totalInsertCount++;

##      mirrorPrint(sprintf("inserting from %d %d / %d\n", $minid, $minpktRef->[$Iid], $totalInsertCount));


      insertIntoTable() if ($cumInsertCount >= $RingSize);
    }
  }
}

sub insertIntoTable {
    $cumulativeInserts =~ s/^,//;  #remove the leading comma
    $query = "INSERT INTO $tableName (".(join(",",@outfieldName)).") VALUES".$cumulativeInserts;
    doQueryUse(\$query);

    $cumulativeInserts = "";
    $cumInsertCount = 0;

    #printHistogram() if rand() < 0.001;
}

sub readSomeMore {
    my ($tableid, $tableVars, $packetStreams, $lfactors, $detailsid) = @_;

    $query = "SELECT * FROM $inputTables[$tableid] WHERE
                   id > $$tableVars[$TV_STARTID] AND id <= $$tableVars[$TV_ENDID]  
                   $zErrConstraint
               ORDER BY id LIMIT $RingSize";
    my $sth = doQuery(\$query);

    my $newmacs = "";
    my $insertIndex;


    while (my $row = $sth->fetchrow_arrayref) {

      my $mtime = $$row[$IbMACTime];
      $mtime += $$detailsid[$DIrollovers];  	##fix wraparounds

      if ($mtime < $$detailsid[$DIlastmtime]) {
	  mirrorPrint(sprintf("rolling over $tableid ($$row[$Iid]): mtime=%d lastmtime=$$detailsid[$DIlastmtime] ro=%d cmtime=$mtime\n", 
			    $mtime - $$detailsid[$DIrollovers], 
			      $$detailsid[$DIrollovers]/$MAX_MTIME));
	  $mtime += $MAX_MTIME;
	  $$detailsid[$DIrollovers]+= $MAX_MTIME;
	  $$detailsid[$DIeffectives]--;
	  
	  ##protection against multiple rollovers
	  ##too bad that we have to put this here; its needed only for kalahari traces
	  while (abs($$row[$IaFrameTime]*$MILLION - $mtime - $$detailsid[$DIstartdelta]) > $ROLLOVER_BOUND_CHECK) {
	      mirrorPrint(sprintf("rolling again $tableid ($$row[$Iid]): mtime=%d lastmtime=$$detailsid[$DIlastmtime] ro=%d cmtime=$mtime\n", 
				  $mtime - $$detailsid[$DIrollovers], 
				  $$detailsid[$DIrollovers]/$MAX_MTIME));
	      $mtime += $MAX_MTIME;
	      $$detailsid[$DIrollovers]+= $MAX_MTIME;
	      $$detailsid[$DIeffectives]--;

	      ## if we get ahead of ftime, something has gone wrong
	      die sprintf("ERROR: cannot rollover more $tableid\n") if ( $mtime > $$row[$IaFrameTime]*$MILLION );
	  }
      }
      ## we missed enough packets since the last one that we need to check for rollovers even when mtime > lastmtime	  
      ## this is still not foolproof but the complete test (in the while condition) is too costly to be run for all packets
      elsif ($mtime - $$detailsid[$DIlastmtime] > $MILLION) {
	  while (abs($$row[$IaFrameTime]*$MILLION - $mtime - $$detailsid[$DIstartdelta]) > $ROLLOVER_BOUND_CHECK) {
	      mirrorPrint(sprintf("rolling again2 $tableid ($$row[$Iid]): mtime=%d lastmtime=$$detailsid[$DIlastmtime] ro=%d cmtime=$mtime\n", 
				  $mtime - $$detailsid[$DIrollovers], 
				  $$detailsid[$DIrollovers]/$MAX_MTIME));
	      $mtime += $MAX_MTIME;
	      $$detailsid[$DIrollovers]+= $MAX_MTIME;
	      $$detailsid[$DIeffectives]--;

	      ## if we get ahead of ftime, something has gone wrong
	      die sprintf("ERROR: cannot rollover more $tableid\n") if ( $mtime > $$row[$IaFrameTime]*$MILLION );	  
	  }
      }

      $$detailsid[$DIlastmtime] = $mtime;

      ##correct time
	##mirrorPrint("$$row[$IbMACTime]");
      $$row[$IbMACTime] = $$lfactors{base} + ($mtime - $$lfactors{base}) * $$lfactors{slope} + $$lfactors{intercept};
	##mirrorPrint(" -> $$row[$IbMACTime]\n");

      ##print STDERR "$tableid: $$lfactors{slope} $$lfactors{intercept}\n";

      ##MACIndices and extra fields
      foreach (@addrFieldsIndices) {
	next if (!defined $$row[$_]);

	my $macaddr = $inputMACIndices[$tableid]->{$$row[$_]};
	if (!defined $macIndices{$macaddr}) {
	    die "ERROR: stable mac entry not found for $macaddr\n" if ($StableMacs);
	    $macIndices{$macaddr} = (1 + keys %macIndices);
	    $newmacs .= ",('$macaddr')";
	}
	$$row[$_] = $macIndices{$macaddr};
      }

      
      ##figure out  the right stream
      my $streamid = $rawTableOffset[$tableid];  ##start by assuming raw
      if (defined($$row[$IzzFromTableIndex])) {      ##correct if there is need
	  $streamid = -1 + $$row[$IzzFromTableIndex] + $rawTableOffset[$tableid];
      }

      ##insert the packet if there is space
      if ($$packetStreams[$streamid]->[$PS_COUNT] < $RingSize) {
	  $insertIndex = ($$packetStreams[$streamid]->[$PS_COUNT] + $$packetStreams[$streamid]->[$PS_INDEX]) % $RingSize;
	  $$packetStreams[$streamid]->[$PS_BUFFER]->[$insertIndex] = [@{$row}];

##	  mirrorPrint(" -> st=$streamid ii $insertIndex id=", $$packetStreams[$streamid]->[$PS_INDEX], " count=", $$packetStreams[$streamid]->[$PS_COUNT], "\n");
	  $$packetStreams[$streamid]->[$PS_COUNT]++;

	  $$tableVars[$TV_STARTID] = $$row[$Iid];
	  $$tableVars[$TV_LASTTIME] = $$row[$IaFrameTime];

      }
      else { 
	  last;  ##one of the streams is full; so quit
      }
  }


    if ($newmacs ne "") {
	$newmacs =~ s/^,//;
	$query = "INSERT INTO $tableNameMac (value) VALUES $newmacs";
	doQueryUse(\$query);
    }
}


sub printHistogram {
    foreach my $key (keys %MergingGap) {
	mirrorPrint("$key merging gap histogram\n");
	my $total = 0;
	foreach my $key1 (sort {$a <=> $b} keys %{$MergingGap{$key}}) {
	    mirrorPrint("$key $key1 ", $MergingGap{$key}->{$key1}, "\n");
	    $total += $MergingGap{$key}->{$key1};
	}
	mirrorPrint("$key total $total\n");
    }
}

sub readInputMACIndices {
    foreach my $id (0..$#inputTables) {
	my $macIndexTable = "$inputTables[$id]_MACIndex";
	$macIndexTable =~ s/debug_//g;   ##in case i am still playing with beaconIndex-ratul2.pl..
	$query = "SELECT id, value FROM $macIndexTable";
	my $sth = doQueryUse(\$query);
	while (my ($mid, $value) = $sth->fetchrow_array) {
	    $inputMACIndices[$id]->{$mid} = $value;
	    $macIndices{$value} = $mid if ($StableMacs and $id == 0);
	}
    }
}

sub initializeBoundaryContinueState {
    ## get mac addresses
    my $query = "SELECT * FROM $tableNameMac";
    my $sth = doQuery(\$query);
    while (my ($id, $value) = $sth->fetchrow_array) {
	$macIndices{$value} = $id;
    }
}

sub initializeContinueState {

    ## get mac addresses
    my $query = "SELECT * FROM $tableNameMac";
    my $sth = doQuery(\$query);
    while (my ($id, $value) = $sth->fetchrow_array) {
	$macIndices{$value} = $id;
    }

    ## get startIds
    $query = "SELECT ".(join(',', map {"MAX(zzid_$_)"} (0..$#inputTables))) . "from $tableName";
    $sth = doQuery(\$query);
    my @starts = $sth->fetchrow_array;
    mirrorPrint("starting points @starts\n");

    ##populate lastBeaconRow
    $query = "SELECT * from $beaconsTable WHERE id_0 <= $starts[0] AND $idConstraint ORDER BY id DESC LIMIT 1";
    $sth = doQuery(\$query);
    my $row = $sth->fetchrow_hashref;
    foreach my $id (0..$#inputTables) {
	$lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[$id]] = $$row{"id_$id"}; 
	$lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]] = $$row{"ftime_$id"};
	$lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] = $$row{"mtime_$id"};

	$details[$id]->[$DIrollovers] = 0;

	while (abs($lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION -
		   $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] - $details[$id]->[$DIstartdelta]) > (2**31)) {
	    $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $MAX_MTIME;
	    $details[$id]->[$DIrollovers]+=$MAX_MTIME;
	}
	$details[$id]->[$DIlastmtime] = $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];

	mirrorPrint("rollover-$id: ", $details[$id]->[$DIrollovers]/$MAX_MTIME,"\n");
    }

    return @starts;
}

sub initializeDetails {
    foreach my $id (0..$#inputTables) {
	my $query = "SELECT aFrameTime, bMACTime from $inputTables[$id] where id > $startIds{$id} limit 1";
	my $sth = doQuery(\$query);
	my ($ftime, $mtime) = $sth->fetchrow_array();

	$details[$id]->[$DIstartdelta] = $ftime*$MILLION - $mtime;
	$details[$id]->[$DIrollovers] = 0;
	$details[$id]->[$DIlastmtime] = 0;
	$details[$id]->[$DIeffectives] = 0;
    }
}

sub insertStableMacs {
    my $query = "INSERT INTO $tableNameMac SELECT * from $inputTables[0]_MACIndex"; 
    doQuery(\$query);
}

sub sqr {
    $_[0] * $_[0];
}


sub computeLinearRegression {
    my ($sth, $lfactors, $query) = @_;

    my ($n, $sumx, $sumx2);
    my (@sumxy, @sumy, @sumy2);

    my @rollovers;
    map {$rollovers[$_] = 0;} (0..$#inputTables);

    my $row;
    while (($row = $sth->fetchrow_arrayref)) {

	##lets not do close beacons
	##next if ($lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_idX]->[0]] != $startIds{0} &&  ## always do the first row
	##	 $row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[0]] - $lastBeaconRow->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[0]] < $BEACON_SKIP_SECS );

	for my $id (0..$#inputTables) {
	    my $originalMtime =       $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];

	    ##$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $details[$id]->[$DIrollovers];
	    $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $rollovers[$id];
	
	    ##die "ERROR: couldn't account for all the effectives for $id\n" if ($details[$id]->[$DIeffectives] != 0);
	
	    ##we might have multiply rolled over since the last beacon; compute the effective mtime
	    ## based on that the starting delta and the current delta should be similar
	    while (abs($row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION -
		       $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] - $details[$id]->[$DIstartdelta])
		   > $ROLLOVER_BOUND_CHECK) {
		my $derails = "id=$id".
		    " beid=".$row->[$beaconEntryIndex[$BEACONENTRY_idX]->[$id]].
		    " omtime=".$originalMtime.
		    " mtime=".$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]].
		    " ro=".$rollovers[$id]/$MAX_MTIME.
		    " ftime=".sprintf("%.7f",$row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION).
		    " delta=".$details[$id]->[$DIstartdelta];
		mirrorPrint("effective $derails\n");
	    
		$rollovers[$id] += $MAX_MTIME;
		$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $MAX_MTIME;
		
		die sprintf("ERROR: cannot get effective mtime. ftime = %.7f; mtime = %Ld\n", 
			    $row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]],
			    $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]])
		    if ( $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] > 1 + $row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION);
		
		$details[$id]->[$DIeffectives]++;
	    }
	}
    
	$n++;
	my $x = $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]];                                                               
	$sumx  += $x;
	$sumx2 += $x * $x;                                                               # compute sum of x**2
	
	foreach my $id (1..$#inputTables) {

	    my $y = $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	    $sumxy[$id] += $x * $y;                                                               # compute sum of x * y
	    $sumy[$id]  += $y;                                                                    # compute sum of y
	    $sumy2[$id] += $y * $y;                                                               # compute sum of y**2
	}
    
    }

    foreach my $id (1..$#inputTables) {
	$$lfactors[$id]->{base} = 0;

	$$lfactors[$id]->{slope} = ($n * $sumxy[$id]  -  $sumx * $sumy[$id]) / ($n * $sumx2 - sqr($sumx));                  # compute slope;
	#$$lfactors[$id]->{intercept} = -1 * (($sumy[$id] * $sumx2  -  $sumx * $sumxy[$id]) / ($n * $sumx2  -  sqr($sumx)));            # compute y-intercept
	
	$$lfactors[$id]->{intercept} = -1 * ($sumy[$id] - $$lfactors[$id]->{slope}* $sumx)/$n;            # compute y-intercept
	
	$$lfactors[$id]->{slope} = 1/$$lfactors[$id]->{slope};
	$$lfactors[$id]->{intercept} = $$lfactors[$id]->{intercept}*$$lfactors[$id]->{slope};

	mirrorPrint("regress lfactors for $id = $$lfactors[$id]->{slope} / $$lfactors[$id]->{intercept}\n");

    }

    ## now, display the residuals
    $sth = doQueryConnUse($conn, \$query);
    map {$rollovers[$_] = 0;} (0..$#inputTables);

    my @residuals;
    while (my $row = $sth->fetchrow_arrayref) {
	for my $id (0..$#inputTables) {
	    my $originalMtime =       $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	    $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $rollovers[$id];
	    
	    ##we might have multiply rolled over since the last beacon; compute the effective mtime
	    ## based on that the starting delta and the current delta should be similar
	    while (abs($row->[$beaconEntryIndex[$BEACONENTRY_ftimeX]->[$id]]*$MILLION -
		       $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] - $details[$id]->[$DIstartdelta])
		   > $ROLLOVER_BOUND_CHECK) {
		$rollovers[$id] += $MAX_MTIME;
		$row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]] += $MAX_MTIME;
	    }
	}
	
	my $x = $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[0]];                                                               
	foreach my $id (1..$#inputTables) {
	    my $y = $row->[$beaconEntryIndex[$BEACONENTRY_mtimeX]->[$id]];
	    my $yprime = $$lfactors[$id]->{intercept} + $y*$$lfactors[$id]->{slope};
	    my $error = $x - $yprime;

	    my $beid = $row->[$BEACONENTRY_id];

	    $residuals[$id] += $error * $error;

	    ##print STDERR "beid=$beid x=$x y=$y yprime=$yprime error=$error\n";
	}
    }

    foreach my $id (1..$#inputTables) {
	mirrorPrint("residual ", $id, " ", $residuals[$id], " ", $n, "\n");
    }

}

sub shortenTableName {
    my $longname = $_[0];

    
    my %shortNames = ("chihuahuan" => "chi", 
		      "sahara" => "sah",
		      "sonoran" => "son",
		      "mojave" => "moj",
		      "kalahari" => "kal");
    
    my $shortname = $longname;
    foreach my $lname (keys %shortNames) {
        $shortname =~ s/$lname/$shortNames{$lname}/g;
    }

    return $shortname;
}
