
## should really be a package ....

use Time::Local;
use DBI;

my $dbUser = (defined $ENV{WDB_USER})? $ENV{WDB_USER} : "default_user";
my $dbPasswd = (defined $ENV{WDB_PASSWD})? $ENV{WDB_PASSWD} : "default_password";
my $dbHost = (defined $ENV{WDB_HOST})? $ENV{WDB_HOST} : "default_host";
my $dbName = (defined $ENV{WDB_DB})? $ENV{WDB_DB} : "default_db";

## use a socket if the db is on the local host
if ($ENV{HOSTNAME} =~ /^$dbHost/) {
    ## if the mysql was not installed in the default place as root
    my $socket = ($dbHost =~ m/^galaxy/)? ":mysql_socket=/galaxy6/bdata2/mysql/mysql.sock" : "";
    $dbHost = "localhost$socket";
}

my $dbh;
my $verbose=0;
my $LOGFILE;

####################################################################
# now()
####################################################################

sub now {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $mon += 1;
    $year += 1900;
    return sprintf("$mon/$mday/$year %2d:%02d:%02d", $hour, $min, $sec);
}

####################################################################
# toUnixTime()
####################################################################

sub toUnixTime{
# argument passed is a date/time string.

    my($time_str)=@_;
    my $unixTime;

    die "You didn't supply a time in call to toUnixTime()" if !defined($time_str);

    # Here are some arrays of conversion stuff.
    # week day conversion:
    my(%MON)=(     "Jan",0,
		   "Feb",1,
		   "Mar",2,
		   "Apr",3,
		   "May",4,
		   "Jun",5,
		   "Jul",6,
		   "Aug",7,
		   "Sep",8,
		   "Oct",9,
		   "Nov",10,
		   "Dec",11);

    foreach (keys %MON) { $MON{"$_,"} = $MON{$_}; }

    # Allow multiple formats.  We assume the caller isn't trying to make this fail...
    my ($day, $month, $year, $hour, $min, $sec );

    my @fields = split(' ',$time_str);
    for (my $index=0; $index<=$#fields; $index++ ) {
      my $f = $fields[$index];

      if ( $f =~ m/^(\d+):(\d+):([\d\.]+)$/ ) {
	$hour = $1;
	$min = $2;
	$sec = $3;
      }
      elsif ( $f =~ /^(\d+)\/(\d+)\/(\d+)$/ ) {
	$mon = $1;
	$day = $2;
	$year = $3;
      }
      elsif ( defined $MON{$f} ) {
	$mon = $MON{$f};
      }
      elsif ( $f =~ m/\d+/ ) {
	if ( $f > 31 or $f eq '04' ) {
	  $year = $f;
	} else {
	  $day = $f;
	}
      }
    }

    $year += 2000 if $year < 100;

    $year-=1900;
    $day =~ s/,$//;

    $unixTime = $sec + timelocal(0, $min, $hour, $day, $mon, $year);
    return $unixTime;
}

####################################################################
# connect/disconnect
####################################################################

sub dbconnect {
    $dbh = DBI->connect( "DBI:mysql:$dbName;host=$dbHost", $dbUser, $dbPasswd) ||
        die "Can't connect to $dbName on $dbHost";
}

sub dbdisconnect {
    $dbh->disconnect;
}

# (caller must disconnect $conn itself)
sub getConnection {
  my $conn = DBI->connect( "DBI:mysql:$dbName;host=$dbHost", $dbUser, $dbPasswd) ||
        die "Can't connect to $dbName on $dbHost";;
  return $conn;
}

####################################################################
# mirrorPrint()
#   Prints to stdout and to $LOGFILE
####################################################################

sub setLogFile {
    $LOGFILE = $_[0];
}

sub mirrorPrint {
    my $arg;
    foreach $arg (@_) {
	print $arg;
    }
    foreach $arg (@_) {
	print $LOGFILE $arg;
    }
}

###############################
# doQuery() and prepareQuery()
###############################

sub setVerbose {
    $verbose = $_[0];
}

sub doQuery {
    my $q = $_[0];
    print "QUERY: '$$q'\n" if ( $verbose );
    my $sth = $dbh->prepare( $$q );
    my $rv = $sth->execute();
    die "Error - '$$q'\n" . $dbh->errstr . "\n"  if ! defined $rv;
    return $sth;
}

sub doQueryUse {
    my $q = $_[0];
    print "QUERY: '$$q'\n" if ( $verbose );
    my $sth = $dbh->prepare( $$q, { "mysql_use_result" => 1} );
    my $rv = $sth->execute();
    die "Error - '$$q'\n" . $dbh->errstr . "\n"  if ! defined $rv;
    return $sth;
}

sub doQueryConnUse {
    my ($conn, $q) = @_;
    print "QUERY: '$$q'\n" if ( $verbose );
    my $sth = $conn->prepare( $$q, { "mysql_use_result" => 1} );
    my $rv = $sth->execute();
    die "Error - '$$q'\n" . $dbh->errstr . "\n"  if ! defined $rv;
    return $sth;
}

# intended for queries like 'SELECT foo, bar FROM table WHERE bar=?'
# followed by $sth->execute("frambus");
# (You may have to get a dedicated connection for this -- call getConnection().)
sub prepareQuery {
  my ($conn, $q) = @_;
  print "QUERY: '$$q'\n" if ( $verbose );
  my $sth = $conn->prepare( $$q );
  return $sth;
}

sub executeQuery {
  my ($conn, $q) = @_;
  print "QUERY: '$$q'\n" if ( $verbose );
  my $sth = $conn->prepare( $$q );
  my $rv = $sth->execute();
  die "Error - '$$q'\n" . $dbh->errstr . "\n"  if ! defined $rv;
  return $sth;
}


############################################
# commify()
############################################

sub commify {
    my $text = reverse $_[0];
    $text =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
    return scalar reverse $text;
}

#####
# setUser(
####

sub setUser {
    $dbUser = $_[0];
}

sub setDB {
    $dbName = $_[0];
}

1;
